-------------------------------------------------------------------------------
-- Localized Lua globals
-------------------------------------------------------------------------------
local _G = getfenv(0)

local string = _G.string
local format = string.format
local strsplit = string.split

local math = _G.math
local floor = math.floor

local table = _G.table
local wipe = table.wipe
local tinsert = table.insert
local tremove = table.remove

local ipairs, pairs = _G.ipairs, _G.pairs

-------------------------------------------------------------------------------
-- Localized Blizzard API
-------------------------------------------------------------------------------
local UnitClass, UnitLevel, UnitName = _G.UnitClass, _G.UnitLevel, _G.UnitName
local GetItemInfo, GetItemGem = _G.GetItemInfo, _G.GetItemGem
local GetInventoryItemLink = _G.GetInventoryItemLink

-------------------------------------------------------------------------------
-- AddOn namespace
-------------------------------------------------------------------------------
local NAME = "GearGauge"
local GearGauge = LibStub("AceAddon-3.0"):NewAddon(NAME, "AceEvent-3.0", "AceConsole-3.0")

local LDB = LibStub("LibDataBroker-1.1")
local QTC = LibStub("LibQTipClick-1.1")
local LBC = LibStub("LibBabble-Class-3.0"):GetLookupTable()
local LIB = LibStub("LibItemBonus-2.0")

local dev = false
--[===[@debug@
dev = true
--@end-debug@]===]
local L = LibStub("AceLocale-3.0"):GetLocale(NAME, "enUS", true, dev)

local DataObj
local tooltip

-------------------------------------------------------------------------------
-- Constants
-------------------------------------------------------------------------------
local TBC_MAXLEV = 60
local WOTLK_MAXLEV = 71
local IS_ENCHANTER = false

local COLOR_RED = "|cffff0000"

local EQUIPMENT = {
	["Head"]		= { minlev = 30,	name = _G["INVTYPE_HEAD"],		mult = 0.8125 },
	["Neck"]		= { minlev = 30,	name = _G["INVTYPE_NECK"],		mult = 0.55 },
	["Shoulder"]		= { minlev = 20,	name = _G["INVTYPE_SHOULDER"],		mult = 0.75 },
	["Shirt"]		= { minlev = 1,		name = _G["INVTYPE_BODY"],		mult = 0 },
	["Chest"]		= { minlev = 1,		name = _G["INVTYPE_CHEST"],		mult = 1 },
	["Waist"]		= { minlev = 1,		name = _G["INVTYPE_WAIST"],		mult = 0.5625 },
	["Legs"]		= { minlev = 1,		name = _G["INVTYPE_LEGS"],		mult = 0.875 },
	["Feet"]		= { minlev = 1,		name = _G["INVTYPE_FEET"],		mult = 0.6875 },
	["Wrist"]		= { minlev = 1,		name = _G["INVTYPE_WRIST"],		mult = 0.4375 },
	["Hands"]		= { minlev = 1,		name = _G["INVTYPE_HAND"],		mult = 0.625 },
	["Finger0"]		= { minlev = 20,	name = _G["INVTYPE_FINGER"],		mult = 0.55 },
	["Finger1"]		= { minlev = 20,	name = _G["INVTYPE_FINGER"],		mult = 0.55 },
	["Trinket0"]		= { minlev = 50,	name = _G["INVTYPE_TRINKET"],		mult = 0.75 },
	["Trinket1"]		= { minlev = 50,	name = _G["INVTYPE_TRINKET"],		mult = 0.75 },
	["Back"]		= { minlev = 1,		name = _G["INVTYPE_CLOAK"],		mult = 0.48 },
	["MainHand"]		= { minlev = 1,		name = _G["INVTYPE_WEAPONMAINHAND"],	mult = 0.50 },
	["SecondaryHand"]	= { minlev = 1,		name = _G["INVTYPE_WEAPONOFFHAND"],	mult = 0.50 },
	["Ranged"]		= { minlev = 1,		name = _G["INVTYPE_RANGED"],		mult = 0.35 },
	["Tabard"]		= { minlev = 1,		name = _G["INVTYPE_TABARD"],		mult = 0 },
}

local ORDERED_EQUIPMENT = {
	[1]	= "Head",	[2]	= "Neck",		[3]	= "Shoulder",
	[4]	= "Shirt",	[5]	= "Chest",		[6]	= "Waist",
	[7]	= "Legs",	[8]	= "Feet",		[9]	= "Wrist",
	[10]	= "Hands",	[11]	= "Finger0",		[12]	= "Finger1",
	[13]	= "Trinket0",	[14]	= "Trinket1",		[15]	= "Back",
	[16]	= "MainHand",	[17]	= "SecondaryHand",	[18]	= "Ranged",
	[19]	= "Tabard"
}

local TWO_HANDED_WEAPON = {
	[L["Fishing Poles"]] = true,
	[L["Polearms"]] = true,
	[L["Staves"]] = true,
	[L["Two-Handed Axes"]] = true,
	[L["Two-Handed Maces"]] = true,
	[L["Two-Handed Swords"]] = true,
}

local GEM_TYPES = {
	"Blue",
	"Meta",
	"Red",
	"Yellow"
}

-------------------------------------------------------------------------------
-- Configuration options (future)
-------------------------------------------------------------------------------
local MAX_LEVEL = 80
local MIN_LEVEL = 1

-------------------------------------------------------------------------------
-- Variables
-------------------------------------------------------------------------------
local char_equip = {}		-- Character's equipment. Generated when AddOn is loaded or state is changed.
local target_equip = {}		-- Target's equipment. Generated when AddOn is loaded or state is changed.

-------------------------------------------------------------------------------
-- Table cache
-------------------------------------------------------------------------------
local AcquireTable, ReleaseTable
do
	local table_cache = {}

	-- Returns a table
	function AcquireTable()
		local tbl = tremove(table_cache) or {}
		return tbl
	end

	-- Cleans the table and stores it in the cache
	function ReleaseTable(tbl)
		if not tbl then return end
		wipe(tbl)
		tinsert(table_cache, tbl)
	end
end	-- do

-------------------------------------------------------------------------------
-- Local functions
-------------------------------------------------------------------------------
local function AverageItemLevels(unit)
	unit = unit or "player"

	local unit_level = UnitLevel(unit)
	local level_cnt = 0
	local item_cnt = 0


	for slot in pairs(EQUIPMENT) do
		if unit == "player" then
			local item = char_equip[slot]

			if item.level > 0 then
				level_cnt = level_cnt + item.level
				item_cnt = item_cnt + 1
			end
		else
			local link = GetInventoryItemLink(unit, EQUIPMENT[slot].id)

			if link then
				local _, _, _, item_level, _, _, _, _, _, _ = GetItemInfo(link)

				if item_level > 0 then
					level_cnt = level_cnt + item_level
					item_cnt = item_cnt + 1
				end
			end
		end
	end
	if item_cnt == 0 then return 0 end

	local avg = floor(level_cnt / item_cnt)
	return avg
end

-------------------------------------------------------------------------------
-- Item status generation
-------------------------------------------------------------------------------
local BuildGearData
do
	local prev_stype	-- Placeholder in case previous slot contained a two-handed weapon.

	-------------------------------------------------------------------------------
	-- Check to see if an empty slot is OK given the unit's level/class.
	-------------------------------------------------------------------------------
	local function IsEmptySlotOK(slot, level, class)
		if slot == "SecondaryHand"  then
			if level < 5 and (class == LBC["Paladin"] or class == LBC["Shaman"] or class == LBC["Warrior"]) then	-- Shields
				return true
			else
				return false
			end
		end

		if slot == "Ranged" then
			if  level < 52 and (class == LBC["Druid"] or class == LBC["Paladin"] or class == LBC["Shaman"]) then	-- Idols, Librams, Totems
				return true
			elseif level < 5 and (class == LBC["Mage"] or class == LBC["Priest"] or class == LBC["Warlock"]) then	-- Wands
				return true
			else
				return false
			end
		end
		return EQUIPMENT[slot].minlev > level
	end

	local function GetEmptySockets(item)
		if not item then return end

		if not item.link or not item.info or not item.info.bonuses then
			item.emptyBlue, item.emptyMeta, item.emptyRed, item.emptyYellow = 0, 0, 0, 0
			return
		end
		local bonuses = item.info.bonuses

		item.emptyBlue = LIB:GetBonus("EMPTY_SOCKET_BLUE", bonuses) or 0
		item.emptyMeta = LIB:GetBonus("EMPTY_SOCKET_META", bonuses) or 0
		item.emptyRed = LIB:GetBonus("EMPTY_SOCKET_RED", bonuses) or 0
		item.emptyYellow = LIB:GetBonus("EMPTY_SOCKET_YELLOW", bonuses) or 0
	end

	-- Item subtypes which are not enchantable
	local NON_ENCHANTABLE = {
		[L["Idols"]] = true,
		[L["Librams"]] = true,
		[L["Miscellaneous"]] = true,
		[L["Other"]] = true,
		[L["Sigils"]] = true,
		[L["Thrown"]] = true,
		[L["Totems"]] = true,
		[L["Wands"]] = true,
	}

	-------------------------------------------------------------------------------
	-- Assign values to an item for later use.
	-------------------------------------------------------------------------------
	local function ConstructItem(item_link, slot, unit)
		local unit_level = UnitLevel(unit)
		local item = AcquireTable()

		-- Default values here
		item.slot, item.rating, item.level = slot, 0, 0
		item.emptyMeta, item.emptyRed, item.emptyYellow, item.emptyBlue = 0, 0, 0, 0
		item.ench_id = -1

		if not item_link then
			local two_handed = slot == "SecondaryHand" and TWO_HANDED_WEAPON[prev_stype]
			local is_ok = IsEmptySlotOK(slot, unit_level, UnitClass(unit))

			if not two_handed and not is_ok then
				item.rating = - EQUIPMENT[slot].mult
			end
			item.empty_ok = is_ok
			prev_stype = nil
			return item
		end
		item.name, item.link, item.rarity, item.level, item.minlev, item.type, item.stype, item.stacksize, item.equiploc, item.texture = GetItemInfo(item_link)
		prev_stype = item.stype

		-- Set the level of Heirloom items depending on the player's level
		if item.rarity == 7 then
			if unit_level <= TBC_MAXLEV then
				item.level = unit_level
			elseif unit_level > TBC_MAXLEV and unit_level <= WOTLK_MAXLEV then
				item.level = 3 * (unit_level - 64) + 90
			elseif unit_level > WOTLK_MAXLEV then
				item.level = 4 * (unit_level - 73) + 142
			end
		end
		if item.level <= 1 or slot == "Tabard" then return item end

		item.info = LIB:ScanItemLink(item_link)

		local points = 0

                if item.info.bonuses then
                        for k, v in pairs(item.info.bonuses) do
				local bonus = LIB:GetBonus(k, item.info.bonuses, unit_level)

				if bonus < 1 then GearGauge:Print("No bonus for "..item.name) else
					points = points + bonus
				end
                        end
                end
		local mult = EQUIPMENT[slot].mult

		if TWO_HANDED_WEAPON[item.stype] then
			mult = mult + mult
		end
		points = ((points + (item.level / unit_level)) * 0.1) * mult

		local rating = points
		local _, _, ench_id = strsplit(":", item_link)		-- Grab the item's enchant id, if enchanted.

		if slot ~= "Waist" and slot ~= "Shirt" and slot ~= "Tabard" then
			item.ench_id = ench_id

			if ench_id == nil or ench_id == "0" then
				if ((slot == "Finger0" or slot == "Finger1") and IS_ENCHANTER) or not NON_ENCHANTABLE[item.stype] then
					rating = rating - (points * 0.25)
				else
					item.ench_id = -1
				end
			end
		end

		-- Get gem information
		GetEmptySockets(item)

		for k, v in ipairs(GEM_TYPES) do
			if item["empty"..v] > 0 then
				rating = rating - ((points * 0.125) * item["empty"..v])
			end
		end
		item.rating = rating

		return item
	end

	-------------------------------------------------------------------------------
	-- List of items equipped by unit.
	-------------------------------------------------------------------------------
	function BuildGearData(unit, item_list)
		for idx, slot in ipairs(ORDERED_EQUIPMENT) do
			ReleaseTable(item_list[slot])
			wipe(item_list)
		end
		local rating = 0

		prev_stype = nil

		for idx, slot in ipairs(ORDERED_EQUIPMENT) do
			local link = GetInventoryItemLink(unit, EQUIPMENT[slot].id)
			local item = ConstructItem(link, slot, unit)

			rating = rating + item.rating
			item_list[slot] = item
		end
		return rating
	end
end -- do

do
	-- Table used to gather information about a target, to be used elsewhere.
	local target = {}

	function GearGauge:Scan(unit)
		if InCombatLockdown() or not CheckInteractDistance("target", 1)
			or not (UnitExists(unit) and UnitIsPlayer(unit)
				and (UnitLevel(unit) >= MIN_LEVEL) and (UnitLevel(unit) <= MAX_LEVEL)) then return end

		if not CanInspect(unit, false) then
			GearGauge:Print(COLOR_RED..target.name..": "..ERR_PLAYER_WRONG_FACTION.."|r")
			return
		end
		NotifyInspect(unit)
		wipe(target)

		-------------------------------------------------------------------------------
		-- For future implementation
		-------------------------------------------------------------------------------
		--		target.name = UnitName(unit)
		--		target.level = UnitLevel(unit)
		--		target.class = UnitClass(unit)
		--		target.gender = UnitSex(unit)
		--		target.race = UnitRace(unit)
		--		target.guild = GetGuildInfo(unit) or "("..NOT_APPLICABLE..")"
		target.faction = UnitFactionGroup(unit)
		--		target.updated = time()

		target.item_list = AcquireTable()
		target.rating = BuildGearData(unit, target.item_list)

		-- Wipe the target's equipment list
		for idx, slot in ipairs(ORDERED_EQUIPMENT) do
			ReleaseTable(target.item_list[slot])
		end
		ReleaseTable(target.item_list)

		local equip_level = AverageItemLevels(unit)
		local unit_name = UnitName(unit)

		if unit_name ~= UnitName("player") then
			local color_str = ""
			if target.faction ~= UnitFactionGroup("player") then color_str = COLOR_RED end

			self:Print(color_str..unit_name.." "..RATING..": "..format("%.2f", target.rating)..". "..ITEMS.." "..LEVEL..": "..equip_level)
		end
	end
end -- do

-------------------------------------------------------------------------------
-- Event handlers
-------------------------------------------------------------------------------
local function UpdateData()
	DataObj.text = format("%.2f", BuildGearData("player", char_equip))
end

function GearGauge:PLAYER_LEVEL_UP()
	UpdateData()
end

function GearGauge:PLAYER_TARGET_CHANGED()
	self:Scan("target")
end

function GearGauge:UNIT_INVENTORY_CHANGED()
	UpdateData()
end

-------------------------------------------------------------------------------
-- Tooltip functions
-------------------------------------------------------------------------------
local GetItemStatus
do
	-- Returns a string based on whether or not the item is enchanted
	function GetItemStatus(item)
		if not item.link then return HELPFRAME_ITEM_TITLE end

		local ench_id = item.ench_id
		local retstr, comma = "", ""

		if ench_id ~= -1 and (ench_id == nil or ench_id == "0") then
			retstr = retstr.."|TInterface\\Icons\\Trade_Engraving:25:25|t"
		end
		local socket_tex = "|TInterface\\ItemSocketingFrame\\UI-EmptySocket-"

		for k, v in ipairs(GEM_TYPES) do
			if item["empty"..v] > 0 then
				retstr = retstr..socket_tex..v..":25:25|tx"..item["empty"..v]
				comma = KEY_COMMA.." "
			end
		end
		return retstr
	end
end --do

local ItemOnEnter, ItemOnLeave
do
	local QTip = LibStub("LibQTip-1.0")
	local bonus_tip

	function ItemOnEnter(event, cell, arg)
		if not arg then return end
		QTC.OnEnter(event, cell, arg)

		bonus_tip = QTip:Acquire(NAME.."BonusTip", 2, "LEFT", "LEFT")
		bonus_tip:ClearAllPoints()
		bonus_tip:SetPoint("TOPRIGHT", tooltip, "TOPLEFT")
		bonus_tip:Clear()

		for k, v in pairs(arg) do
			bonus_tip:AddLine(k, v)
		end

		if TipTac and TipTac.AddModifiedTip then TipTac:AddModifiedTip(bonus_tip, true) end

		bonus_tip:Show()
	end

	function ItemOnLeave(event, cell, arg)
		if not bonus_tip then return end
		QTC.OnLeave(event, cell, arg)
		bonus_tip = QTip:Release(bonus_tip)
	end
end	-- do

local function BuildTooltip(anchor, unit)
	unit = unit or "player"

	tooltip = QTC:Acquire(NAME.."Tooltip", 3, "LEFT", "LEFT", "LEFT")
	tooltip:SetCallback("OnEnter", ItemOnEnter)
	tooltip:SetCallback("OnLeave", ItemOnLeave)

	tooltip:Clear()
	tooltip:SetAutoHideDelay(0.25, anchor)
	tooltip:SmartAnchorTo(anchor)

	if TipTac and TipTac.AddModifiedTip then TipTac:AddModifiedTip(tooltip, true) end

	local rating_tot = 0
	local equip_level = AverageItemLevels(unit)
	local line, column

	line, column = tooltip:AddNormalHeader()
	tooltip:SetNormalCell(line, column, LEVEL..": "..UnitLevel(unit)..KEY_COMMA.." "..ITEMS.." "..LEVEL..": "..equip_level, "CENTER", 3)
	tooltip:AddSeparator()
	line, column = tooltip:AddNormalHeader("", RATING, ADDON_MISSING)

	tooltip:SetNormalCell(line, 1, ITEMS, "CENTER")
	tooltip:AddSeparator()

	local two_handed = nil
	local prev_stype = nil

	for idx, slot in ipairs(ORDERED_EQUIPMENT) do
		local item = char_equip[slot]

		rating_tot = rating_tot + item.rating

		if item.link then
			local r, g, b, hex = GetItemQualityColor(item.rarity)

			line, column = tooltip:AddNormalLine()
			tooltip:SetLineColor(line, r, g, b)

			if not item.info or not item.info.bonuses then
				tooltip:SetNormalCell(line, 1, "|T"..item.texture..":25:25|t"..item.name)
			else
				tooltip:SetCell(line, 1, "|T"..item.texture..":25:25|t"..item.name, item.info.bonuses)
			end
			tooltip:SetNormalCell(line, 2, format("%.2f", item.rating))
			tooltip:SetNormalCell(line, 3, GetItemStatus(item))
			prev_stype = item.stype
		else
			local two_handed = slot == "SecondaryHand" and TWO_HANDED_WEAPON[prev_stype]

			if not two_handed  then
				line, column = tooltip:AddLine()

				if item.empty_ok then
					tooltip:SetLineColor(line)
				else
					tooltip:SetLineColor(line, 1)
				end
				tooltip:SetNormalCell(line, 1, EQUIPMENT[slot].name, nil, nil, 1, nil, 20)
				tooltip:SetNormalCell(line, 2, format("%.2f", item.rating))
				tooltip:SetNormalCell(line, 3, GetItemStatus(item))
				prev_stype = nil
			end
		end
	end
	tooltip:AddSeparator()

	line, column = tooltip:AddLine()

	tooltip:SetNormalCell(line, 2, format("%.2f", rating_tot))
	tooltip:Show()
end

local function ShowTooltip(self)
	BuildTooltip(self)
end

-------------------------------------------------------------------------------
-- Initialization functions
-------------------------------------------------------------------------------
function GearGauge:OnInitialize()
	-- Set the numeric ID and texture strings for the equipment slots
	for slot in pairs(EQUIPMENT) do
		EQUIPMENT[slot].id, EQUIPMENT[slot].texture = GetInventorySlotInfo(slot.."Slot")
	end

	-- Determine if the toon is an enchanter
	local enchanting_sklname = GetSpellInfo(7411)
	for idx = 1, GetNumSkillLines() do
		local skill = GetSkillLineInfo(idx)
		if skill == enchanting_sklname then
			IS_ENCHANTER = true
			break
		end
	end

	DataObj = LDB:NewDataObject("GearGauge", {
		type	= "data source",
		label	= "GearGauge",
		text	= "0",
		icon	= "Interface\\Icons\\INV_Shield_36",
		OnEnter = function(self) ShowTooltip(self) end,
		OnLeave = function() end,
	})
	self.OnInitialize = nil
end

function GearGauge:OnEnable()
	UpdateData()
	GearGauge:RegisterEvent("PLAYER_LEVEL_UP")
	GearGauge:RegisterEvent("PLAYER_TARGET_CHANGED")
	GearGauge:RegisterEvent("UNIT_INVENTORY_CHANGED")
end

function GearGauge:OnDisable()
	GearGauge:UnregisterEvent("PLAYER_LEVEL_UP")
	GearGauge:UnregisterEvent("PLAYER_TARGET_CHANGED")
	GearGauge:UnregisterEvent("UNIT_INVENTORY_CHANGED")
end
